#ifndef __ThreadListener_h__
#define __ThreadListener_h__

#include "AutoSocket.h"
#include "ISelector.h"
#include "ISocket.h"
#include "IThreadListener.h"
#include "IWeaver.h"
#include "SyncRequest.h"

class ThreadListener : public IThreadListener
{
public:
    ThreadListener( IWeaver &weaver, ISelector &selector, ISocket &socket ) : _weaver( weaver ),
                                                                              _selector( selector ),
                                                                              _socket( &socket ),
                                                                              _threadWaitingForEvent( false ),
                                                                              _syncPointIndex( 0 )
    {
        registerToSelector();
    }

    ~ThreadListener()
    {
        unregisterFromSelector();
        discardSocket();
    }

    int descriptor()
    {
        return _socket->descriptor();
    }

    void canRead()
    {
        HeapData receivedData;

        readSocketData( receivedData );

        if ( !haveWaitingEvents() )
        {
            informWeaver( receivedData );
        }
        else
        {
            resolveSyncRequest( receivedData );
        }
    }

    void announceChosenEvent( Event &event )
    {
        storeAnnouncedEvent( event );
        if ( threadWaitingForEvent() )
        {
            sendEventToThread( extractEvent() );
        }
    }

private:
    IWeaver &_weaver;
    ISelector &_selector;
    ISocket *_socket;
    Vector<Event> _announcedEvents;
    bool _threadWaitingForEvent;
    unsigned long long _syncPointIndex;

    void registerToSelector()
    {
        _selector.registerToSelector( *this );
    }

    void unregisterFromSelector()
    {
        _selector.unregisterFromSelector( *this );
    }

    void discardSocket()
    {
        T::discardSocket( _socket );
        _socket = 0;
    }

    void readSocketData( HeapData &receivedData )
    {
        _socket->receive( receivedData );
    }

    void informWeaver( HeapData &receivedData )
    {
        incrementSyncPointIndex();
        _threadWaitingForEvent = true;

        SyncRequest syncRequest( receivedData );
        _weaver.bSync( this,
                       syncRequest.getRequested(),
                       syncRequest.getWatched(),
                       syncRequest.getBlocked(),
                       syncRequest.getLabel(),
                       _syncPointIndex );
    }

    void storeAnnouncedEvent( Event &event )
    {
        _announcedEvents.append( event );
    }

    bool haveWaitingEvents() const
    {
        return !_announcedEvents.empty();
    }

    bool threadWaitingForEvent() const
    {
        return _threadWaitingForEvent;
    }

    void sendEventToThread( Event event )
    {
        HeapData serialized;
        event.serialize( serialized );
        _socket->send( serialized );

        _threadWaitingForEvent = false;
    }

    void resolveSyncRequest( HeapData &receivedData )
    {
        bool resolved = false;

        while ( !resolved )
        {
            if ( haveWaitingEvents() )
            {
                incrementSyncPointIndex();
                Event announcedEvent = extractEvent();

                if ( threadRequestedOrWatched( receivedData, announcedEvent ) )
                {
                    sendEventToThread( announcedEvent );
                    resolved = true;
                }
            }
            else
            {
                informWeaver( receivedData );
                resolved = true;
            }
        }
    }

    static bool threadRequestedOrWatched( HeapData &receivedData, Event announcedEvent )
    {
        SyncRequest syncRequest( receivedData );
        return syncRequest.getRequested().exists( announcedEvent ) || syncRequest.getWatched().exists( announcedEvent );
    }

    Event extractEvent()
    {
        return _announcedEvents.popFirst();
    }

    void incrementSyncPointIndex()
    {
        ++_syncPointIndex;
    }
};

#endif // __ThreadListener_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
