#ifndef __ThreadStateGraph_h__
#define __ThreadStateGraph_h__

#include "File.h"
#include "Map.h"
#include "Stringf.h"
#include "ThreadState.h"

struct ThreadStateGraph
{
    Map<String, ThreadState> _graph;
    Vector<Event> _events;
    String _initialState;

    void clear()
    {
        _graph.clear();
        _events.clear();
        _initialState = "";
    }

    void applyAxioms()
    {
        removeNonRequested();
        eliminateNonReachables();
    }

    Vector<String> bfsStateEnumeration()
    {
        init();

        while ( moreStatesToCheck() )
        {
            checkAnotherState();
        }

        addAllUnreachedStates();

        return _bfsEnumeration;
    }

    void dump()
    {
        printf( "State graph size is: %u\n\n", _graph.size() );

        Map<String, ThreadState>::iterator it;
        for ( it = _graph.begin(); it != _graph.end(); ++it )
        {
            printf( "%s:\n", it->second._label.ascii() );

            printf( "Requested: " );
            for ( unsigned i = 0; i < it->second._requestedEvents.size(); ++i )
            {
                printf( "%u ", it->second._requestedEvents[i].type() );
            }
            printf( "\n" );

            printf( "Watched: " );
            for ( unsigned i = 0; i < it->second._watchedEvents.size(); ++i )
            {
                printf( "%u ", it->second._watchedEvents[i].type() );
            }
            printf( "\n" );

            printf( "Blocked: " );
            for ( unsigned i = 0; i < it->second._blockedEvents.size(); ++i )
            {
                printf( "%u ", it->second._blockedEvents[i].type() );
            }
            printf( "\n" );

            printf( "History: " );
            for ( unsigned i = 0; i < it->second._history.size(); ++i )
            {
                printf( "%u (%u) ", it->second._history[i]._event.type(),
                        it->second._history[i]._environmentConfiguration );
            }
            printf( "\n" );

            printf( "Successors: " );
            for ( unsigned i = 0; i < it->second._successors.size(); ++i )
            {
                printf( "%u-->%s  ", it->second._successors[i]._event.type(), it->second._successors[i]._state.ascii() );
            }
            printf( "\n" );

            printf( "Predecessors: " );
            for ( unsigned i = 0; i < it->second._predecessors.size(); ++i )
            {
                printf( "%u-->%s  ", it->second._predecessors[i]._event.type(),
                        it->second._predecessors[i]._state.ascii() );
            }
            printf( "\n" );

            printf( "Atomic Propositions: " );
            for ( unsigned i = 0; i < it->second._atomicPropositions.size(); ++i )
            {
                printf( "%s ", it->second._atomicPropositions.at( i ).ascii() );
            }
            printf( "\n\n" );
        }
    }

    void dumpToFile( const String &path, const String &className )
    {
        File output( path );
        output.open( IFile::MODE_WRITE );

        createHeader( output, className );
        createFunction( output );
        createFooter( output, className );
    }

private:
    Vector<String> _bfsEnumeration;
    Vector<String> _statesToCheck;

    void init()
    {
        _bfsEnumeration.clear();
        _statesToCheck.clear();

        _statesToCheck.append( _initialState );
        _bfsEnumeration.append( _initialState );
    }

    bool moreStatesToCheck() const
    {
        return !_statesToCheck.empty();
    }

    void checkAnotherState()
    {
        Vector<String> successors = _graph[_statesToCheck.popFirst()].allSuccessors();

        for ( unsigned i = 0; i < successors.size(); ++i )
        {
            if ( !_bfsEnumeration.exists( successors[i] ) )
            {
                _bfsEnumeration.append( successors[i] );
                _statesToCheck.append( successors[i] );
            }
        }
    }

    void addAllUnreachedStates()
    {
        Map<String, ThreadState>::iterator it;

        for ( it = _graph.begin(); it != _graph.end(); ++it )
        {
            if ( !_bfsEnumeration.exists( it->first ) )
            {
                _bfsEnumeration.append( it->first );
            }
        }
    }

    void removeNonRequested()
    {
        Map<String, ThreadState>::iterator it;

        for ( it = _graph.begin(); it != _graph.end(); ++it )
        {
            removeNonRequstedAtState( it->first );
        }
    }

    void removeNonRequstedAtState( const String &label )
    {
        unsigned i = 0;

        while ( i < _graph[label]._successors.size() )
        {
            if ( !_graph[label]._requestedEvents.exists( _graph[label]._successors[i]._event ) )
            {
                removeTransition( label, i );
            }
            else
            {
                ++i;
            }
        }
    }

    void removeTransition( const String &label, unsigned successorIndex )
    {
        String successor = _graph[label]._successors[successorIndex]._state;

        Transition predecessor( _graph[label]._successors[successorIndex]._event, label );
        _graph[successor]._predecessors.erase( predecessor );

        _graph[label]._successors.eraseAt( successorIndex );
    }

    void eliminateNonReachables()
    {
        init();

        while ( moreStatesToCheck() )
        {
            checkAnotherState();
        }

        Vector<String> toErase;

        Map<String, ThreadState>::iterator it;

        for ( it = _graph.begin(); it != _graph.end(); ++it )
        {
            if ( !_bfsEnumeration.exists( it->first ) )
            {
                toErase.append( it->first );
            }
        }

        for ( unsigned i = 0; i < toErase.size(); ++i )
        {
            removeState( toErase[i] );
        }
    }

    void removeState( const String &label )
    {
        for ( unsigned i = 0; i < _graph[label]._successors.size(); ++i )
        {
            String successor = _graph[label]._successors[i]._state;
            _graph[successor].removePredecessor( label );
        }

        for ( unsigned i = 0; i < _graph[label]._predecessors.size(); ++i )
        {
            String predecessor = _graph[label]._predecessors[i]._state;
            _graph[predecessor].removeSuccessor( label );
        }

        _graph.erase( label );
    }

    void createHeader( File &output, const String &className )
    {
        output.write( Stringf( "#ifndef __%s_h__\n", className.ascii() ) );
        output.write( Stringf( "#define __%s_h__\n", className.ascii() ) );
        output.write( "\n" );
        output.write( "#include \"ThreadStateGraph.h\"" );
        output.write( "\n" );
        output.write( "\n" );
        output.write( Stringf( "class %s\n", className.ascii() ) );
        output.write( "{\n" );
        output.write( "public:\n" );
        output.write( "\n" );
    }

    void createFooter( File &output, const String &className )
    {
        output.write( "};\n" );
        output.write( "\n" );
        output.write( Stringf( "#endif // __%s_h__", className.ascii() ) );
    }

    void createFunction( File &output )
    {
        output.write( "\tvoid getGraph( ThreadStateGraph &threadStateGraph )\n" );
        output.write( "\t{\n" );

        for ( unsigned i = 0; i < _events.size(); ++i )
        {
            output.write( Stringf( "\t\tthreadStateGraph._events.append( %u );\n", _events[i].type() ) );
        }

        output.write( "\n" );

        output.write( Stringf( "\t\tthreadStateGraph._initialState = \"%s\";\n\n", _initialState.ascii() ) );

        Map<String, ThreadState>::iterator it = _graph.begin();
        for ( unsigned i = 0; i < _graph.size(); ++i )
        {
            output.write( Stringf( "\t\tThreadState s%u;\n", i ) );
            output.write( Stringf( "\t\ts%u._label = \"%s\";\n", i, it->second._label.ascii() ) );

            for ( unsigned j = 0; j < it->second._requestedEvents.size(); ++j )
            {
                output.write( Stringf( "\t\ts%u._requestedEvents.append( %u );\n",
                                       i, it->second._requestedEvents[j].type() ) );
            }

            for ( unsigned j = 0; j < it->second._watchedEvents.size(); ++j )
            {
                output.write( Stringf( "\t\ts%u._watchedEvents.append( %u );\n",
                                       i, it->second._watchedEvents[j].type() ) );
            }

            for ( unsigned j = 0; j < it->second._blockedEvents.size(); ++j )
            {
                output.write( Stringf( "\t\ts%u._blockedEvents.append( %u );\n",
                                       i, it->second._blockedEvents[j].type() ) );
            }

            for ( unsigned j = 0; j < it->second._successors.size(); ++j )
            {
                output.write( Stringf( "\t\ts%u._successors.append( Transition( %u, \"%s\" ) );\n",
                                       i,
                                       it->second._successors[j]._event.type(),
                                       it->second._successors[j]._state.ascii() ) );
            }

            output.write( Stringf( "\t\tthreadStateGraph._graph[\"%s\"] = s%u;\n\n", it->second._label.ascii(), i ) );

            ++it;
        }

        output.write( "\t}\n" );
    }
};

#endif // __ThreadStateGraph_h__

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
