#ifndef __Weaver_h__
#define __Weaver_h__

#include "AutoThreadListener.h"
#include "CoreMessage.h"
#include "Event.h"
#include "EventSelector.h"
#include "EventSets.h"
#include "IEngine.h"
#include "ISocket.h"
#include "IWeaver.h"
#include "Map.h"
#include "ThreadStateGraph.h"
#include "UniqueVector.h"
#include "Vector.h"

class Weaver : public IWeaver
{
public:
    Weaver( ISelector &selector, IEngine &engine ) : _selector( selector ), _engine( engine ), _syncPointIndex( 0 )
    {
    }

    ~Weaver()
    {
        discardListeners();
    }

    void addLocalStateGraph( const String &threadName, ThreadStateGraph &threadStateGraph )
    {
        _localThreadsToStateGraphs[threadName] = &threadStateGraph;
        _threadsToLocationInGraph[threadName] = threadStateGraph._initialState;
    }

    void addForeignStateGraph( const String &threadName, ThreadStateGraph &threadStateGraph )
    {
        _foreignThreadsToStateGraphs[threadName] = &threadStateGraph;
        _threadsToLocationInGraph[threadName] = threadStateGraph._initialState;
    }

	void handleAcceptedSocket( ISocket &socket )
	{
        IThreadListener *threadListener = T::createThreadListener( *this, _selector, socket );
        storeThreadName( socket, threadListener );
        storeListener( threadListener );

        if ( allLocalsAccountedFor() )
        {
            printf( "Weaver: all accounted for, starting\n" );
            fflush( 0 );

            attemptEventSelection();
        }
    }

    void incomingCoreMessage( CoreMessage &coreMessage )
    {
        storeCoreMessage( coreMessage );
        resolveCoreMessages();
        attemptEventSelection();
    }

    void bSync( IThreadListener *thread,
                const Vector<Event> &,
                const Vector<Event> &,
                const Vector<Event> &,
                const String &label,
                unsigned long long synchronizationIndex )
    {
        storeBSync( thread, label, synchronizationIndex - 1 );
        resolveSyncMessages();
        attemptEventSelection();
    }

    void storeBSync( IThreadListener *thread, const String &label, unsigned long long synchronizationIndex )
    {
        String threadName = _listenerssToNames[thread];

        printf( "Weaver: Incoming bsync message: %llu -- %s in %s\n", synchronizationIndex,
                threadName.ascii(), label.ascii() );
        fflush( 0 );

        CoreMessage coreMessage( synchronizationIndex, threadName, label );

        _bSyncMessages.append( coreMessage );
    }

private:
    ISelector &_selector;
    IEngine &_engine;
    Vector<IThreadListener *> _threadListeners;
    UniqueVector<Event> _requested;
    UniqueVector<Event> _blocked;
    Event _selectedEvent;

    Map<String, ThreadStateGraph *> _localThreadsToStateGraphs;
    Map<String, ThreadStateGraph *> _foreignThreadsToStateGraphs;

    Map<String, String> _threadsToLocationInGraph;
    Map<IThreadListener *, String> _listenerssToNames;

    Vector<String> _threadsInUnknownState;
    unsigned long long _syncPointIndex;
    Vector<CoreMessage> _coreMessages;
    Vector<CoreMessage> _bSyncMessages;

    void attemptEventSelection()
    {
        printf( "Weaver: attemping selection\n" );
        fflush( 0 );

        while ( !haveThreadsInUnknownState() )
        {
            incrementSyncIndex();
            selectNextEvent();
            resolveCoreMessages();
            resolveSyncMessages();
        }

        printf( "Weaver: threads in unknown state:\n" );
        fflush( 0 );

        for ( unsigned i = 0; i < _threadsInUnknownState.size(); ++i )
        {
            printf( "\t%s\n", _threadsInUnknownState[i].ascii() );
            fflush( 0 );
        }
    }

    void storeCoreMessage( CoreMessage &coreMessage )
    {
        printf( "Weaver: Incoming core message: %llu -- %s in %s\n", coreMessage.synchronizationIndex(),
                coreMessage.threadName().ascii(), coreMessage.chosenLabel().ascii() );
        fflush( 0 );
        _coreMessages.append( coreMessage );
    }

    void incrementSyncIndex()
    {
        ++_syncPointIndex;
    }

    bool allLocalsAccountedFor() const
    {
        return _threadListeners.size() == _localThreadsToStateGraphs.size();
    }

    bool haveThreadsInUnknownState() const
    {
        return !_threadsInUnknownState.empty();
    }

    void selectNextEvent()
    {
        selectEvent();
        announceEvent();
        clearDataStructures();
    }

    void storeListener( IThreadListener *threadListener )
    {
        _threadListeners.append( threadListener );
    }

    void storeThreadName( ISocket &socket, IThreadListener *threadListener )
    {
        HeapData nameData;
        socket.receive( nameData );
        _listenerssToNames[threadListener] = String( (char *)nameData.data(), nameData.size() );
    }

    void discardListeners()
    {
        for ( unsigned i = 0; i < _threadListeners.size(); ++i )
        {
            T::discardThreadListener( _threadListeners[i] );
        }
    }

    void collectRequested()
    {
        Map<String, ThreadStateGraph *>::iterator it;

        for ( it = _localThreadsToStateGraphs.begin(); it != _localThreadsToStateGraphs.end(); ++it )
        {
            String state = _threadsToLocationInGraph[it->first];
            _requested += it->second->_graph[state]._requestedEvents;
        }

        for ( it = _foreignThreadsToStateGraphs.begin(); it != _foreignThreadsToStateGraphs.end(); ++it )
        {
            String state = _threadsToLocationInGraph[it->first];
            _requested += it->second->_graph[state]._requestedEvents;
        }
    }

    void collectBlocked()
    {
        Map<String, ThreadStateGraph *>::iterator it;

        for ( it = _localThreadsToStateGraphs.begin(); it != _localThreadsToStateGraphs.end(); ++it )
        {
            String state = _threadsToLocationInGraph[it->first];
            _blocked += it->second->_graph[state]._blockedEvents;
        }

        for ( it = _foreignThreadsToStateGraphs.begin(); it != _foreignThreadsToStateGraphs.end(); ++it )
        {
            String state = _threadsToLocationInGraph[it->first];
            _blocked += it->second->_graph[state]._blockedEvents;
        }
    }

    void selectEvent()
    {
        collectRequested();
        collectBlocked();
        invokeEventChoose();
    }

    void invokeEventChoose()
    {
        _selectedEvent = EventSelector::select( _requested, _blocked );
    }

    bool notBlocked( Event event )
    {
        return !_blocked.exists( event );
    }

    void announceEvent()
    {
        printf( "Weaver: announcing #%llu is %u\n", _syncPointIndex, _selectedEvent.type() );
        fflush( 0 );

        announceToLocalThreads();
        updateStateGraphs();
    }

    void announceToLocalThreads()
    {
        for ( unsigned i = 0; i < _threadListeners.size(); ++i )
        {
            _threadListeners[i]->announceChosenEvent( _selectedEvent );
        }
    }

    void clearDataStructures()
    {
        _requested.clear();
        _blocked.clear();
    }

    void updateStateGraphs()
    {
        Map<String, ThreadStateGraph *>::iterator it;

        for ( it = _localThreadsToStateGraphs.begin(); it != _localThreadsToStateGraphs.end(); ++it )
        {
            updateStateGraph( it->first, it->second, _selectedEvent );
        }

        for ( it = _foreignThreadsToStateGraphs.begin(); it != _foreignThreadsToStateGraphs.end(); ++it )
        {
            updateStateGraph( it->first, it->second, _selectedEvent );
        }
    }

    void updateStateGraph( String threadName, ThreadStateGraph *threadStateGraph, Event &event )
    {
        String currentStateLabel = _threadsToLocationInGraph[threadName];
        ThreadState *currentState = &threadStateGraph->_graph[currentStateLabel];

        if ( currentState->_requestedEvents.exists( event ) || currentState->_watchedEvents.exists( event ) )
        {
            if ( currentState->deterministicEvent( event ) )
            {
                String newStateLabel = currentState->labelOfSuccessor( event );
                _threadsToLocationInGraph[threadName] = newStateLabel;
            }
            else
            {
                printf( "\t\tWeaver: Thread in underterministic state\n" );
                fflush( 0 );

                _threadsInUnknownState.append( threadName );
            }
        }
    }

    void resolveCoreMessages()
    {
        Vector<CoreMessage> copy = _coreMessages;

        for ( unsigned i = 0; i < copy.size(); ++i )
        {
            if ( copy[i].synchronizationIndex() == _syncPointIndex )
            {
                _coreMessages.eraseByValue( copy[i] );
                _threadsToLocationInGraph[copy[i].threadName()] = copy[i].chosenLabel();
                _threadsInUnknownState.eraseByValue( copy[i].threadName() );
            }
            else if ( copy[i].synchronizationIndex() < _syncPointIndex )
            {
                _coreMessages.eraseByValue( copy[i] );
            }
        }
    }

    void resolveSyncMessages()
    {
        Vector<CoreMessage> copy = _bSyncMessages;

        for ( unsigned i = 0; i < copy.size(); ++i )
        {
            if ( copy[i].synchronizationIndex() == _syncPointIndex )
            {
                if ( _threadsInUnknownState.exists( copy[i].threadName() ) )
                {
                    _engine.sendCoreMessage( copy[i] );
                    _bSyncMessages.eraseByValue( copy[i] );
                    _threadsToLocationInGraph[copy[i].threadName()] = copy[i].chosenLabel();
                    _threadsInUnknownState.eraseByValue( copy[i].threadName() );
                }
                else
                {
                    _bSyncMessages.eraseByValue( copy[i] );
                }
            }
            else if ( copy[i].synchronizationIndex() < _syncPointIndex )
            {
                _bSyncMessages.eraseByValue( copy[i] );
            }
        }
    }
};

#endif // __Weaver_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
