#include <cxxtest/TestSuite.h>

#include "BProgram.h"
#include "Event.h"
#include "MockBThread.h"
#include "MockEngineFactory.h"
#include "MockErrno.h"
#include "MockParameters.h"

class MockForBProgram :
    public MockErrno,
    public MockEngineFactory
{
public:
};

class BProgramTestSuite : public CxxTest::TestSuite
{
public:
	MockForBProgram *mock;
    MockEngine *engine;
    MockParameters *parameters;

	void setUp()
	{
		TS_ASSERT( mock = new MockForBProgram );
		TS_ASSERT( parameters = new MockParameters );

        engine = &( mock->mockEngine );
	}

	void tearDown()
	{
		TS_ASSERT_THROWS_NOTHING( delete mock );
	}

    void test_constructor()
    {
        BProgram *bProgram;

        TS_ASSERT( bProgram = new BProgram( "/home/config.ini" ) );

        TS_ASSERT_EQUALS( parameters->lastLoadedPath, "/home/config.ini" );

        TS_ASSERT_THROWS_NOTHING( delete bProgram );
    }

    void test_run_program()
    {
        BProgram *bProgram;

        TS_ASSERT( bProgram = new BProgram( "dontCare" ) );

        MockBThread thread1;
        MockBThread thread2;
        MockBThread thread3;

        TS_ASSERT_THROWS_NOTHING( bProgram->addThread( thread1 ) );
        TS_ASSERT_THROWS_NOTHING( bProgram->addThread( thread2 ) );
        TS_ASSERT_THROWS_NOTHING( bProgram->addThread( thread3 ) );

        TS_ASSERT_THROWS_NOTHING( bProgram->runProgram() );

        TS_ASSERT( !engine->runWasCalled );

        TS_ASSERT( thread1.runWasCalled );
        TS_ASSERT( thread2.runWasCalled );
        TS_ASSERT( thread3.runWasCalled );

        TS_ASSERT_THROWS_NOTHING( delete bProgram );

        TS_ASSERT( engine->wasDiscarded );
    }

    void test_run_engine()
    {
        BProgram *bProgram;

        TS_ASSERT( bProgram = new BProgram( "dontCare" ) );

        TS_ASSERT( engine->wasCreated );

        TS_ASSERT_THROWS_NOTHING( bProgram->runEngine() );

        TS_ASSERT( engine->runWasCalled );

        TS_ASSERT_THROWS_NOTHING( delete bProgram );

        TS_ASSERT( engine->wasDiscarded );
    }

    void test_add_local_blocking_data()
    {
        BProgram bProgram( "dontCare" );

        ThreadStateGraph threadStateGraph;

        threadStateGraph._events.append( Event( 1 ) );
        threadStateGraph._events.append( Event( 2 ) );

        threadStateGraph._initialState = "a";

        ThreadState a;
        ThreadState b;

        a._label = "a";
        a._requestedEvents.append( Event( 1 ) );
        a._blockedEvents.append( Event( 2 ) );
        a._successors.append( Transition( Event( 1 ), "b" ) );

        b._label = "b";
        b._watchedEvents.append( Event( 2 ) );
        b._successors.append( Transition( Event( 2 ), "b" ) );
        b._predecessors.append( Transition( Event( 1 ), "a" ) );
        b._history.append( HistoryItem( Event( 1 ), 0 ) );

        threadStateGraph._graph["a"] = a;
        threadStateGraph._graph["b"] = b;

        TS_ASSERT_THROWS_NOTHING( bProgram.addLocalStateGraph( "Jefferson", threadStateGraph ) );

        TS_ASSERT_EQUALS( engine->lastThreadName, "Jefferson" );
        TS_ASSERT_EQUALS( engine->lastLocalStateGraph, &threadStateGraph );
    }

    void test_add_foreign_blocking_data()
    {
        BProgram bProgram( "dontCare" );

        ThreadStateGraph threadStateGraph;

        threadStateGraph._events.append( Event( 1 ) );
        threadStateGraph._events.append( Event( 2 ) );

        threadStateGraph._initialState = "a";

        ThreadState a;
        ThreadState b;

        a._label = "a";
        a._requestedEvents.append( Event( 1 ) );
        a._blockedEvents.append( Event( 2 ) );
        a._successors.append( Transition( Event( 1 ), "b" ) );

        b._label = "b";
        b._watchedEvents.append( Event( 2 ) );
        b._successors.append( Transition( Event( 2 ), "b" ) );
        b._predecessors.append( Transition( Event( 1 ), "a" ) );
        b._history.append( HistoryItem( Event( 1 ), 0 ) );

        threadStateGraph._graph["a"] = a;
        threadStateGraph._graph["b"] = b;

        TS_ASSERT_THROWS_NOTHING( bProgram.addForeignStateGraph( "Jefferson", threadStateGraph ) );

        TS_ASSERT_EQUALS( engine->lastThreadName, "Jefferson" );
        TS_ASSERT_EQUALS( engine->lastForeignStateGraph, &threadStateGraph );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
