#include <cxxtest/TestSuite.h>

#include "CoreCommunicator.h"
#include "EngineConfiguration.h"
#include "MockErrno.h"
#include "MockParameters.h"
#include "MockSelector.h"
#include "MockServerSocketFactory.h"
#include "MockSocketFactory.h"
#include "MockWeaver.h"
#include "RealMalloc.h"

class MockForCoreCommunicator :
    public MockErrno,
    public RealMalloc,
    public MockServerSocketFactory,
    public MockSocketFactory
{
public:
};

class CoreCommunicatorTestSuite : public CxxTest::TestSuite
{
public:
	MockForCoreCommunicator *mock;
    MockParameters *parameters;
	MockServerSocket *serverSocket;
    MockSelector *selector;
    MockSocket *socket;
    MockWeaver *weaver;

	void setUp()
	{
		TS_ASSERT( mock = new MockForCoreCommunicator );
		TS_ASSERT( parameters = new MockParameters );
        TS_ASSERT( selector = new MockSelector );
        TS_ASSERT( weaver = new MockWeaver );

		serverSocket = &( mock->mockServerSocket );
        socket = &( mock->mockSocket );
    }

	void tearDown()
	{
        TS_ASSERT_THROWS_NOTHING( delete weaver );
        TS_ASSERT_THROWS_NOTHING( delete selector );
		TS_ASSERT_THROWS_NOTHING( delete parameters );
        TS_ASSERT_THROWS_NOTHING( delete mock );
    }

    void test_core_one()
    {
        parameters->nextCoreNumber = 1;
        parameters->nextCoreCommunicationPort = 6655;

        CoreCommunicator *coreCommunicator;

        TS_ASSERT( coreCommunicator = new CoreCommunicator( *selector, *weaver ) );

        TS_ASSERT( !socket->wasCreatedIpPort );

		TS_ASSERT( serverSocket->wasCreated );
		TS_ASSERT_EQUALS( serverSocket->lastSelector, selector );
		TS_ASSERT_EQUALS( serverSocket->lastPort, 6655U );
		TS_ASSERT_EQUALS( serverSocket->lastAccepter, coreCommunicator );

        MockSocket mockSocket;
        mockSocket.mockConstructor( 17 );

        TS_ASSERT( !selector->registerWasCalled );

        TS_ASSERT_THROWS_NOTHING( coreCommunicator->handleAcceptedSocket( mockSocket ) );

        TS_ASSERT( selector->registerWasCalled );
        TS_ASSERT_EQUALS( selector->lastRegisteredSelectable, coreCommunicator );
        TS_ASSERT_EQUALS( selector->lastRegistetedDescriptor, 17 );

        TS_ASSERT_EQUALS( coreCommunicator->descriptor(), 17 );

		TS_ASSERT( serverSocket->wasDiscarded );

        class SpecialDiscarder : public T::Base_discardSocket
        {
        public:
            ISocket *lastDiscardedSocket;

            void discardSocket( ISocket *socket )
            {
                lastDiscardedSocket = socket;
            }
        } specialDiscarder;

        TS_ASSERT( !selector->unregisterWasCalled );

        TS_ASSERT_THROWS_NOTHING( delete coreCommunicator );

        TS_ASSERT( selector->unregisterWasCalled );
        TS_ASSERT_EQUALS( selector->lastUnregisteredSelectable, coreCommunicator );
        TS_ASSERT_EQUALS( selector->lastUnregistetedDescriptor, 17 );

        TS_ASSERT_EQUALS( specialDiscarder.lastDiscardedSocket, &mockSocket );
    }

    void test_constructor_core_two()
    {
        parameters->nextCoreNumber = 2;
        parameters->nextCoreCommunicationPort = 6655;
        socket->lastDescriptor = 43;

        CoreCommunicator *coreCommunicator;

        TS_ASSERT( coreCommunicator = new CoreCommunicator( *selector, *weaver ) );

        TS_ASSERT( socket->wasCreatedIpPort );
        TS_ASSERT_EQUALS( socket->lastPort, 6655U );
        TS_ASSERT_EQUALS( socket->lastIp, inet_addr( EngineConfiguration::LOOPBACK_IP ) );

        TS_ASSERT( selector->registerWasCalled );
        TS_ASSERT_EQUALS( selector->lastRegisteredSelectable, coreCommunicator );
        TS_ASSERT_EQUALS( selector->lastRegistetedDescriptor, 43 );

        TS_ASSERT_EQUALS( coreCommunicator->descriptor(), 43 );

        TS_ASSERT( !serverSocket->wasCreated );

        TS_ASSERT( !selector->unregisterWasCalled );

        TS_ASSERT_THROWS_NOTHING( delete coreCommunicator );

        TS_ASSERT( selector->unregisterWasCalled );
        TS_ASSERT_EQUALS( selector->lastUnregisteredSelectable, coreCommunicator );
        TS_ASSERT_EQUALS( selector->lastUnregistetedDescriptor, 43 );

		TS_ASSERT( socket->wasDiscarded );
    }

    void test_can_read()
    {
        parameters->nextCoreNumber = 2;
        parameters->nextCoreCommunicationPort = 6655;

        unsigned char incomingData[] = {
            0xef, 0xcd, 0xab, 0x89, 0x67, 0x45, 0x23, 0x01,
            0x03, 0x00, 0x00, 0x00,
            0x6a, 0x6f, 0x6e,
            0x0b, 0x00, 0x00, 0x00,
            0x68, 0x65, 0x6c, 0x6c, 0x6f, 0x20, 0x77, 0x6f, 0x72, 0x6c, 0x64
        };

        ConstSimpleData serialized( incomingData, sizeof(incomingData) );
        socket->nextReceivedData.append( serialized );

        CoreCommunicator coreCommunicator( *selector, *weaver );

        TS_ASSERT_THROWS_NOTHING( coreCommunicator.canRead() );

        TS_ASSERT_EQUALS( weaver->lastCoreMessages.size(), 1U );

        TS_ASSERT_EQUALS( weaver->lastCoreMessages[0].synchronizationIndex(), 0x0123456789ABCDEFull );
        TS_ASSERT_EQUALS( weaver->lastCoreMessages[0].threadName(), "jon" );
        TS_ASSERT_EQUALS( weaver->lastCoreMessages[0].chosenLabel(), "hello world" );
    }

    void test_can_read__two_messages()
    {
        parameters->nextCoreNumber = 2;
        parameters->nextCoreCommunicationPort = 6655;

        unsigned char incomingData[] = {
            0xef, 0xcd, 0xab, 0x89, 0x67, 0x45, 0x23, 0x01,
            0x03, 0x00, 0x00, 0x00,
            0x6a, 0x6f, 0x6e,
            0x0b, 0x00, 0x00, 0x00,
            0x68, 0x65, 0x6c, 0x6c, 0x6f, 0x20, 0x77, 0x6f, 0x72, 0x6c, 0x64,

            0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa,
            0x02, 0x00, 0x00, 0x00,
            0x6a, 0x6f,
            0x0b, 0x00, 0x00, 0x00,
            0x68, 0x65, 0x6c, 0x6c, 0x6f, 0x20, 0x77, 0x6f, 0x72, 0x6c, 0x64
        };

        ConstSimpleData serialized( incomingData, sizeof(incomingData) );
        socket->nextReceivedData.append( serialized );

        CoreCommunicator coreCommunicator( *selector, *weaver );

        TS_ASSERT_THROWS_NOTHING( coreCommunicator.canRead() );

        TS_ASSERT_EQUALS( weaver->lastCoreMessages.size(), 2U );

        TS_ASSERT_EQUALS( weaver->lastCoreMessages[0].synchronizationIndex(), 0x0123456789ABCDEFull );
        TS_ASSERT_EQUALS( weaver->lastCoreMessages[0].threadName(), "jon" );
        TS_ASSERT_EQUALS( weaver->lastCoreMessages[0].chosenLabel(), "hello world" );

        TS_ASSERT_EQUALS( weaver->lastCoreMessages[1].synchronizationIndex(), 0xAAAAAAAAAAAAAAAAull );
        TS_ASSERT_EQUALS( weaver->lastCoreMessages[1].threadName(), "jo" );
        TS_ASSERT_EQUALS( weaver->lastCoreMessages[1].chosenLabel(), "hello world" );
    }

    void test_send()
    {
        parameters->nextCoreNumber = 2;
        parameters->nextCoreCommunicationPort = 6655;

        CoreCommunicator coreCommunicator( *selector, *weaver );

        CoreMessage coreMessage( 0x0123456789ABCDEFull, "jon", "hello world" );

        TS_ASSERT_THROWS_NOTHING( coreCommunicator.send( coreMessage ) );

        unsigned char expectedData[] = {
            0xef, 0xcd, 0xab, 0x89, 0x67, 0x45, 0x23, 0x01,
            0x03, 0x00, 0x00, 0x00,
            0x6a, 0x6f, 0x6e,
            0x0b, 0x00, 0x00, 0x00,
            0x68, 0x65, 0x6c, 0x6c, 0x6f, 0x20, 0x77, 0x6f, 0x72, 0x6c, 0x64
        };

        TS_ASSERT_EQUALS( socket->lastSentData.size(), sizeof(expectedData) );
        TS_ASSERT_SAME_DATA( socket->lastSentData.data(), expectedData, sizeof(expectedData) );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
