#include <cxxtest/TestSuite.h>

#include "Engine.h"
#include "MockConfigurationFileFactory.h"
#include "MockCoreCommunicatorFactory.h"
#include "MockParameters.h"
#include "MockSelectorFactory.h"
#include "MockServerSocketFactory.h"
#include "MockTruth.h"
#include "MockWeaverFactory.h"
#include "ThreadStateGraph.h"

class MockForEngine :
    public MockConfigurationFileFactory,
    public MockCoreCommunicatorFactory,
	public MockServerSocketFactory,
	public MockSelectorFactory,
    public MockWeaverFactory,
	public MockTruth
{
public:
};

class EngineTestSuite : public CxxTest::TestSuite
{
public:
	MockForEngine *mock;
	MockServerSocket *serverSocket;
	MockSelector *selector;
	MockWeaver *weaver;
    MockConfigurationFile *configFile;
    MockParameters *parameters;
    MockCoreCommunicator *coreCommunicator;

	void setUp()
	{
		TS_ASSERT( mock = new MockForEngine );
		TS_ASSERT( parameters = new MockParameters );

		serverSocket = &( mock->mockServerSocket );
		selector = &( mock->mockSelector );
		weaver = &( mock->mockWeaver );
        configFile = &( mock->mockConfigurationFile );
        coreCommunicator = &( mock->mockCoreCommunicator );
	}

	void tearDown()
	{
		TS_ASSERT_THROWS_NOTHING( delete mock );
	}

	void test_constructor()
	{
        parameters->nextBThreadManagementPort = 1234;

        Engine *engine;

		TS_ASSERT( engine = new Engine );

		TS_ASSERT( selector->wasCreated );

		TS_ASSERT( weaver->wasCreated );
        TS_ASSERT_EQUALS( weaver->lastSelector, selector );
        TS_ASSERT_EQUALS( weaver->lastEngine, engine );

        TS_ASSERT( coreCommunicator->wasCreated );
        TS_ASSERT_EQUALS( coreCommunicator->lastSelector, selector );
        TS_ASSERT_EQUALS( coreCommunicator->lastWeaver, weaver );

		TS_ASSERT( serverSocket->wasCreated );
		TS_ASSERT_EQUALS( serverSocket->lastSelector, selector );
		TS_ASSERT_EQUALS( serverSocket->lastPort, 1234U );
		TS_ASSERT_EQUALS( serverSocket->lastAccepter, weaver );

		TS_ASSERT_THROWS_NOTHING( delete engine );
		TS_ASSERT( selector->wasDiscarded );
		TS_ASSERT( serverSocket->wasDiscarded );
		TS_ASSERT( weaver->wasDiscarded );
        TS_ASSERT( coreCommunicator->wasDiscarded );
	}

	void test_entry_point()
	{
		mock->numberOfTruths = 5;

		TS_ASSERT_THROWS_NOTHING( Engine::main() );

		TS_ASSERT_EQUALS( selector->numberOfSelectCalls, 5U );
	}

    void test_add_local_blocking_data()
    {
        Engine engine;

        ThreadStateGraph threadStateGraph;

        threadStateGraph._events.append( Event( 1 ) );
        threadStateGraph._events.append( Event( 2 ) );

        threadStateGraph._initialState = "a";

        ThreadState a;
        ThreadState b;

        a._label = "a";
        a._requestedEvents.append( Event( 1 ) );
        a._blockedEvents.append( Event( 2 ) );
        a._successors.append( Transition( Event( 1 ), "b" ) );

        b._label = "b";
        b._watchedEvents.append( Event( 2 ) );
        b._successors.append( Transition( Event( 2 ), "b" ) );
        b._predecessors.append( Transition( Event( 1 ), "a" ) );
        b._history.append( HistoryItem( Event( 1 ), 0 ) );

        threadStateGraph._graph["a"] = a;
        threadStateGraph._graph["b"] = b;

        TS_ASSERT_THROWS_NOTHING( engine.addLocalStateGraph( "Jefferson", threadStateGraph ) );

        TS_ASSERT_EQUALS( weaver->lastThreadName, "Jefferson" );
        TS_ASSERT_EQUALS( weaver->lastLocalStateGraph, &threadStateGraph );
    }

    void test_add_foreign_blocking_data()
    {
        Engine engine;

        ThreadStateGraph threadStateGraph;

        threadStateGraph._events.append( Event( 1 ) );
        threadStateGraph._events.append( Event( 2 ) );

        threadStateGraph._initialState = "a";

        ThreadState a;
        ThreadState b;

        a._label = "a";
        a._requestedEvents.append( Event( 1 ) );
        a._blockedEvents.append( Event( 2 ) );
        a._successors.append( Transition( Event( 1 ), "b" ) );

        b._label = "b";
        b._watchedEvents.append( Event( 2 ) );
        b._successors.append( Transition( Event( 2 ), "b" ) );
        b._predecessors.append( Transition( Event( 1 ), "a" ) );
        b._history.append( HistoryItem( Event( 1 ), 0 ) );

        threadStateGraph._graph["a"] = a;
        threadStateGraph._graph["b"] = b;

        TS_ASSERT_THROWS_NOTHING( engine.addForeignStateGraph( "Jefferson", threadStateGraph ) );

        TS_ASSERT_EQUALS( weaver->lastThreadName, "Jefferson" );
        TS_ASSERT_EQUALS( weaver->lastForeignStateGraph, &threadStateGraph );
    }

    void test_send_core_message()
    {
        Engine engine;

        CoreMessage coreMessage( 0xdd, "jon", "hi there" );
        TS_ASSERT_THROWS_NOTHING( engine.sendCoreMessage( coreMessage ) );

        TS_ASSERT_EQUALS( coreCommunicator->lastSentLabel, "hi there" );
        TS_ASSERT_EQUALS( coreCommunicator->lastSentName, "jon" );
        TS_ASSERT_EQUALS( coreCommunicator->lastSentMessageIndex, 0xddull );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
