#include <cxxtest/TestSuite.h>

#include "MockConfigurationFileFactory.h"
#include "MockErrno.h"
#include "Parameters.h"

class MockForParameters :
    public MockErrno,
    public MockConfigurationFileFactory
{
public:
};

class ParametersTestSuite : public CxxTest::TestSuite
{
public:
    MockForParameters *mock;
    MockConfigurationFile *configFile;

    void setUp()
    {
        TS_ASSERT( mock = new MockForParameters );

        configFile = &( mock->mockConfigurationFile );
    }

    void tearDown()
    {
        TS_ASSERT_THROWS_NOTHING( delete mock );
    }

    void test_load_parameters()
    {
        String path( "/home/config.ini" );

        configFile->keysToValues[ "bthread_management_port" ] = "9997";
        configFile->keysToValues[ "core_number" ] = "1";
        configFile->keysToValues[ "core_communication_port" ] = "3121";

        TS_ASSERT_THROWS_NOTHING( Parameters::loadParameters( path ) );

        TS_ASSERT( configFile->wasCreated );
        TS_ASSERT_EQUALS( configFile->lastPath, path );
        TS_ASSERT( configFile->wasDiscarded );
    }

    void test_get_parameters()
    {
        configFile->keysToValues[ "bthread_management_port" ] = "9997";
        configFile->keysToValues[ "core_number" ] = "1";
        configFile->keysToValues[ "core_communication_port" ] = "3121";

        TS_ASSERT_THROWS_NOTHING( Parameters::loadParameters( "somepath" ) );

        TS_ASSERT_EQUALS( Parameters::bThreadManagementPort(), 9997U );
        TS_ASSERT_EQUALS( Parameters::coreNumber(), 1U );
        TS_ASSERT_EQUALS( Parameters::coreCommunicationPort(), 3121U );
    }


    void test_load_parameters__no_management_port()
    {
        configFile->keysToValues[ "core_number" ] = "1";
        configFile->keysToValues[ "core_communication_port" ] = "3121";

        String path( "/home/config.ini" );

        TS_ASSERT_THROWS_ASSERT( Parameters::loadParameters( path ),
                                 const Error &e,
                                 ( e.code() == Error::VALUE_DOESNT_EXIST_IN_CONFIG_FILE ) &&
                                 ( e.message() == "bthread_management_port" ) );
    }

    void test_load_parameters__no_core_number()
    {
        configFile->keysToValues[ "bthread_management_port" ] = "9997";
        configFile->keysToValues[ "core_communication_port" ] = "3121";

        String path( "/home/config.ini" );

        TS_ASSERT_THROWS_ASSERT( Parameters::loadParameters( path ),
                                 const Error &e,
                                 ( e.code() == Error::VALUE_DOESNT_EXIST_IN_CONFIG_FILE ) &&
                                 ( e.message() == "core_number" ) );
    }

    void test_load_parameters__no_core_communication_port()
    {
        configFile->keysToValues[ "bthread_management_port" ] = "9997";
        configFile->keysToValues[ "core_number" ] = "1";

        String path( "/home/config.ini" );

        TS_ASSERT_THROWS_ASSERT( Parameters::loadParameters( path ),
                                 const Error &e,
                                 ( e.code() == Error::VALUE_DOESNT_EXIST_IN_CONFIG_FILE ) &&
                                 ( e.message() == "core_communication_port" ) );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
