#include <cxxtest/TestSuite.h>

#include "MockErrno.h"
#include "MockSelector.h"
#include "MockSocket.h"
#include "MockSocketFactory.h"
#include "MockWeaver.h"
#include "RealMalloc.h"
#include "ThreadListener.h"

class MockForThreadListener :
    public RealMalloc,
    public MockErrno,
    public MockSocketFactory
{
public:
};

class ThreadListenerTestSuite : public CxxTest::TestSuite
{
public:
	MockForThreadListener *mock;
    MockSelector *selector;
    MockWeaver *weaver;
    MockSocket *socket;

	void setUp()
	{
		TS_ASSERT( mock = new MockForThreadListener );
        TS_ASSERT( selector = new MockSelector );
        TS_ASSERT( weaver = new MockWeaver );

        socket = &( mock->mockSocket );
        socket->wasCreated = true;
    }

	void tearDown()
	{
        TS_ASSERT_THROWS_NOTHING( delete weaver );
        TS_ASSERT_THROWS_NOTHING( delete selector );
        TS_ASSERT_THROWS_NOTHING( delete mock );
	}

    void test_constructor_destructor()
    {
        ThreadListener *threadListener;

        TS_ASSERT( threadListener = new ThreadListener( *weaver, *selector, *socket ) );

        TS_ASSERT_EQUALS( selector->lastRegistetedDescriptor, socket->descriptor() );
        TS_ASSERT_EQUALS( selector->lastRegisteredSelectable, threadListener );

        TS_ASSERT_THROWS_NOTHING( delete threadListener );

        TS_ASSERT_EQUALS( selector->lastUnregisteredSelectable, threadListener );
        TS_ASSERT( socket->wasDiscarded );
    }

    void test_can_read__and_announce()
    {
        ThreadListener threadListener( *weaver, *selector, *socket );

        char serialized[] = {
            // Label
            0x02, 0x00, 0x00, 0x00,
            0x68, 0x69,

            // Requested
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x04, 0x00, 0x00, 0x00, // Length of first
            0x0a, 0x0b, 0x0c, 0x0e, // First

            // WatchedFor
            0x02, 0x00, 0x00, 0x00, // Number of events
            0x04, 0x00, 0x00, 0x00, // Length of first
            0x11, 0x22, 0x33, 0x1a, // First
            0x04, 0x00, 0x00, 0x00, // Length of second
            0x66, 0x54, 0x00, 0x02, // Second

            // Blocked
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x04, 0x00, 0x00, 0x00, // Length of first
            0x33, 0x11, 0x94, 0xa5, // First

            // Atomic Propositions
            0x00, 0x00, 0x00, 0x00 // Number of propositions
        };

        socket->nextReceivedData.append( ConstSimpleData( serialized, sizeof(serialized) ) );

        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );

        TS_ASSERT_EQUALS( weaver->lastRequested.size(), 1U );
        TS_ASSERT_EQUALS( weaver->lastRequested[0].type(), 0x0e0c0b0aU );

        TS_ASSERT_EQUALS( weaver->lastWatched.size(), 2U );
        TS_ASSERT_EQUALS( weaver->lastWatched[0].type(), 0x1a332211U );
        TS_ASSERT_EQUALS( weaver->lastWatched[1].type(), 0x02005466U );

        TS_ASSERT_EQUALS( weaver->lastBlocked.size(), 1U );
        TS_ASSERT_EQUALS( weaver->lastBlocked[0].type(), 0xa5941133U );

        TS_ASSERT_EQUALS( weaver->lastThread, &threadListener );

        TS_ASSERT_EQUALS( weaver->lastLabel, "hi" );

        TS_ASSERT_EQUALS( weaver->lastSynchronizationIndex, 1ull );

        Event announcedEvent( 0x04030201 );
        char serializedEvent[] = { 0x01, 0x02, 0x03, 0x04 };

        TS_ASSERT_THROWS_NOTHING( threadListener.announceChosenEvent( announcedEvent ) );

        TS_ASSERT_EQUALS( socket->lastSentData.size(), sizeof(serializedEvent) );
        TS_ASSERT_SAME_DATA( socket->lastSentData.data(), serializedEvent, sizeof(serializedEvent) );
    }

    void test_annouce_delayed_until_can_read()
    {
        ThreadListener threadListener( *weaver, *selector, *socket );

        Event announcedEvent1( 0x04030201 );
        char serializedEvent1[] = { 0x01, 0x02, 0x03, 0x04 };

        Event announcedEvent2( 0x04030202 );
        char serializedEvent2[] = { 0x02, 0x02, 0x03, 0x04 };

        TS_ASSERT_THROWS_NOTHING( threadListener.announceChosenEvent( announcedEvent1 ) );
        TS_ASSERT_EQUALS( socket->lastSentData.size(), 0U );
        TS_ASSERT_THROWS_NOTHING( threadListener.announceChosenEvent( announcedEvent2 ) );
        TS_ASSERT_EQUALS( socket->lastSentData.size(), 0U );

        char serialized[] = {
            // Label
            0x02, 0x00, 0x00, 0x00,
            0x68, 0x69,

            // Requested
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x04, 0x00, 0x00, 0x00, // Length of first
            0x01, 0x02, 0x03, 0x04, // First

            // WatchedFor
            0x02, 0x00, 0x00, 0x00, // Number of events
            0x04, 0x00, 0x00, 0x00, // Length of first
            0x11, 0x22, 0x33, 0x1a, // First
            0x04, 0x00, 0x00, 0x00, // Length of second
            0x02, 0x02, 0x03, 0x04, // Second

            // Blocked
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x04, 0x00, 0x00, 0x00, // Length of first
            0x33, 0x11, 0x94, 0xa5, // First

            // Atomic Propositions
            0x00, 0x00, 0x00, 0x00 // Number of propositions
        };

        socket->nextReceivedData.append( ConstSimpleData( serialized, sizeof(serialized) ) );
        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );

        TS_ASSERT_EQUALS( socket->lastSentData.size(), sizeof(serializedEvent1) );
        TS_ASSERT_SAME_DATA( socket->lastSentData.data(), serializedEvent1, sizeof(serializedEvent1) );
        TS_ASSERT( !weaver->bSyncWasCalled );

        socket->nextReceivedData.append( ConstSimpleData( serialized, sizeof(serialized) ) );
        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );

        TS_ASSERT_EQUALS( socket->lastSentData.size(), sizeof(serializedEvent2) );
        TS_ASSERT_SAME_DATA( socket->lastSentData.data(), serializedEvent2, sizeof(serializedEvent2) );
        TS_ASSERT( !weaver->bSyncWasCalled );

        socket->nextReceivedData.append( ConstSimpleData( serialized, sizeof(serialized) ) );
        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );

        TS_ASSERT_EQUALS( weaver->lastRequested.size(), 1U );
        TS_ASSERT_EQUALS( weaver->lastRequested[0].type(), 0x04030201U );

        TS_ASSERT_EQUALS( weaver->lastWatched.size(), 2U );
        TS_ASSERT_EQUALS( weaver->lastWatched[0].type(), 0x1a332211U );
        TS_ASSERT_EQUALS( weaver->lastWatched[1].type(), 0x04030202U );

        TS_ASSERT_EQUALS( weaver->lastBlocked.size(), 1U );
        TS_ASSERT_EQUALS( weaver->lastBlocked[0].type(), 0xa5941133U );

        TS_ASSERT_EQUALS( weaver->lastThread, &threadListener );

        TS_ASSERT_EQUALS( weaver->lastLabel, "hi" );

        TS_ASSERT_EQUALS( weaver->lastSynchronizationIndex, 3ull );
    }

    void test_can_read__request_passes_old_announces_and_gets_passed_to_weaver()
    {
        ThreadListener threadListener( *weaver, *selector, *socket );

        Event announcedEvent1( 0x04030201 );
        Event announcedEvent2( 0x04030202 );

        TS_ASSERT_THROWS_NOTHING( threadListener.announceChosenEvent( announcedEvent1 ) );
        TS_ASSERT_EQUALS( socket->lastSentData.size(), 0U );
        TS_ASSERT_THROWS_NOTHING( threadListener.announceChosenEvent( announcedEvent2 ) );
        TS_ASSERT_EQUALS( socket->lastSentData.size(), 0U );

        char serialized[] = {
            // Label
            0x02, 0x00, 0x00, 0x00,
            0x68, 0x69,

            // Requested
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x04, 0x00, 0x00, 0x00, // Length of first
            0x0a, 0x02, 0x03, 0x04, // First

            // WatchedFor
            0x02, 0x00, 0x00, 0x00, // Number of events
            0x04, 0x00, 0x00, 0x00, // Length of first
            0x11, 0x22, 0x33, 0x1a, // First
            0x04, 0x00, 0x00, 0x00, // Length of second
            0xd2, 0x02, 0x03, 0x04, // Second

            // Blocked
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x04, 0x00, 0x00, 0x00, // Length of first
            0x33, 0x11, 0x94, 0xa5, // First

            // Atomic Propositions
            0x00, 0x00, 0x00, 0x00 // Number of propositions
        };

        socket->nextReceivedData.append( ConstSimpleData( serialized, sizeof(serialized) ) );
        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );

        TS_ASSERT_EQUALS( socket->lastSentData.size(), 0U );
        TS_ASSERT( weaver->bSyncWasCalled );

        TS_ASSERT_EQUALS( weaver->lastRequested.size(), 1U );
        TS_ASSERT_EQUALS( weaver->lastRequested[0].type(), 0x0403020aU );

        TS_ASSERT_EQUALS( weaver->lastWatched.size(), 2U );
        TS_ASSERT_EQUALS( weaver->lastWatched[0].type(), 0x1a332211U );
        TS_ASSERT_EQUALS( weaver->lastWatched[1].type(), 0x040302d2U );

        TS_ASSERT_EQUALS( weaver->lastBlocked.size(), 1U );
        TS_ASSERT_EQUALS( weaver->lastBlocked[0].type(), 0xa5941133U );

        TS_ASSERT_EQUALS( weaver->lastLabel, "hi" );

        TS_ASSERT_EQUALS( weaver->lastSynchronizationIndex, 3ull );

        TS_ASSERT_EQUALS( weaver->lastThread, &threadListener );

        Event announcedEvent( 0x040302d2 );
        char serializedEvent[] = { 0xd2, 0x02, 0x03, 0x04 };

        TS_ASSERT_THROWS_NOTHING( threadListener.announceChosenEvent( announcedEvent ) );

        TS_ASSERT_EQUALS( socket->lastSentData.size(), sizeof(serializedEvent) );
        TS_ASSERT_SAME_DATA( socket->lastSentData.data(), serializedEvent, sizeof(serializedEvent) );
    }

    void test_correct_counting_of_sync_points()
    {
        ThreadListener threadListener( *weaver, *selector, *socket );

        Event announcedEvent1( 0x04030201 );
        Event announcedEvent2( 0x04030202 );

        TS_ASSERT_THROWS_NOTHING( threadListener.announceChosenEvent( announcedEvent1 ) );
        TS_ASSERT_THROWS_NOTHING( threadListener.announceChosenEvent( announcedEvent2 ) );

        char serialized[] = {
            // Label
            0x02, 0x00, 0x00, 0x00,
            0x68, 0x69,

            // Requested
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x04, 0x00, 0x00, 0x00, // Length of first
            0x0a, 0x02, 0x03, 0x04, // First

            // WatchedFor
            0x02, 0x00, 0x00, 0x00, // Number of events
            0x04, 0x00, 0x00, 0x00, // Length of first
            0x11, 0x22, 0x33, 0x1a, // First
            0x04, 0x00, 0x00, 0x00, // Length of second
            0xd2, 0x02, 0x03, 0x04, // Second

            // Blocked
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x04, 0x00, 0x00, 0x00, // Length of first
            0x33, 0x11, 0x94, 0xa5, // First

            // Atomic Propositions
            0x00, 0x00, 0x00, 0x00 // Number of propositions
        };

        socket->nextReceivedData.append( ConstSimpleData( serialized, sizeof(serialized) ) );
        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );

        TS_ASSERT_EQUALS( weaver->lastSynchronizationIndex, 3ull );
        TS_ASSERT_THROWS_NOTHING( threadListener.announceChosenEvent( announcedEvent1 ) );

        socket->nextReceivedData.append( ConstSimpleData( serialized, sizeof(serialized) ) );
        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );

        TS_ASSERT_EQUALS( weaver->lastSynchronizationIndex, 4ull );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
