#include <cxxtest/TestSuite.h>

#include "MockEngine.h"
#include "MockErrno.h"
#include "MockSelector.h"
#include "MockSocket.h"
#include "MockThreadListenerFactory.h"
#include "RealMalloc.h"
#include "ThreadStateGraph.h"
#include "Weaver.h"

class MockForWeaver :
    public MockErrno,
    public RealMalloc,
    public MockThreadListenerFactory
{
public:
};

class WeaverTestSuite : public CxxTest::TestSuite
{
public:
	MockForWeaver *mock;
    MockSelector *selector;
    MockEngine *engine;
    MockSocket *socket1;
    MockSocket *socket2;
    MockSocket *socket3;
    MockThreadListener *threadListener1;
    MockThreadListener *threadListener2;
    MockThreadListener *threadListener3;

	void setUp()
	{
		TS_ASSERT( mock = new MockForWeaver );

        TS_ASSERT( selector = new MockSelector );
        TS_ASSERT( engine = new MockEngine );
        TS_ASSERT( socket1 = new MockSocket );
        TS_ASSERT( socket2 = new MockSocket );
        TS_ASSERT( socket3 = new MockSocket );

        setThreadName( socket1, "JonDoe" );
        setThreadName( socket2, "JonDoe" );
        setThreadName( socket3, "JonDoe" );

        threadListener1 = mock->registerThreadListener( socket1 );
        threadListener2 = mock->registerThreadListener( socket2 );
        threadListener3 = mock->registerThreadListener( socket3 );
    }

	void tearDown()
	{
		TS_ASSERT_THROWS_NOTHING( delete socket3 );
        TS_ASSERT_THROWS_NOTHING( delete socket2 );
        TS_ASSERT_THROWS_NOTHING( delete socket1 );
        TS_ASSERT_THROWS_NOTHING( delete engine );
		TS_ASSERT_THROWS_NOTHING( delete selector );
        TS_ASSERT_THROWS_NOTHING( delete mock );
	}

    void setThreadName( MockSocket *socket, String name )
    {
        socket->nextReceivedData.clear();
        socket->nextReceivedData.append( ConstSimpleData( name.ascii(), name.length() ) );
    }

    void test_handle_accepted_socket()
    {
        Weaver *weaver;

        TS_ASSERT( weaver = new Weaver( *selector, *engine ) );

        TS_ASSERT_THROWS_NOTHING( weaver->handleAcceptedSocket( *socket1 ) );

        TS_ASSERT( threadListener1->wasCreated );
        TS_ASSERT_EQUALS( threadListener1->lastWeaver, weaver );
        TS_ASSERT_EQUALS( threadListener1->lastSelector, selector );
        TS_ASSERT_EQUALS( threadListener1->lastSocket, socket1 );

        TS_ASSERT_THROWS_NOTHING( weaver->handleAcceptedSocket( *socket2 ) );

        TS_ASSERT( threadListener2->wasCreated );
        TS_ASSERT_EQUALS( threadListener2->lastWeaver, weaver );
        TS_ASSERT_EQUALS( threadListener2->lastSelector, selector );
        TS_ASSERT_EQUALS( threadListener2->lastSocket, socket2 );

        TS_ASSERT( !threadListener1->wasDiscarded );
        TS_ASSERT( !threadListener2->wasDiscarded );

        TS_ASSERT_THROWS_NOTHING( delete weaver );

        TS_ASSERT( threadListener1->wasDiscarded );
        TS_ASSERT( threadListener2->wasDiscarded );
    }

    void prepareStateGraphOne( ThreadStateGraph &threadStateGraph )
    {
        Event event1( 1 );
        Event event2( 2 );

        threadStateGraph._events.append( event1 );
        threadStateGraph._events.append( event2 );

        threadStateGraph._initialState = "a";

        ThreadState a;
        ThreadState b;

        a._label = "a";
        a._requestedEvents.append( event1 );
        a._successors.append( Transition( event1, "a" ) );
        a._successors.append( Transition( event1, "b" ) );

        b._label = "b";
        b._watchedEvents.append( event1 );
        b._watchedEvents.append( event2 );
        b._successors.append( Transition( event1, "a" ) );
        b._successors.append( Transition( event1, "a" ) );

        threadStateGraph._graph["a"] = a;
        threadStateGraph._graph["b"] = b;
    }

    void prepareStateGraphTwo( ThreadStateGraph &threadStateGraph )
    {
        Event event1( 1 );
        Event event2( 2 );

        threadStateGraph._events.append( event1 );
        threadStateGraph._events.append( event2 );

        threadStateGraph._initialState = "a";

        ThreadState c;
        ThreadState d;
        ThreadState e;

        c._label = "c";
        c._watchedEvents.append( event1 );
        c._successors.append( Transition( event1, "d" ) );
        c._successors.append( Transition( event1, "e" ) );

        d._label = "d";
        d._requestedEvents.append( event1 );
        d._successors.append( Transition( event1, "c" ) );

        e._label = "e";
        e._requestedEvents.append( event2 );
        e._successors.append( Transition( event2, "c" ) );

        threadStateGraph._graph["c"] = c;
        threadStateGraph._graph["d"] = d;
        threadStateGraph._graph["e"] = e;
    }

    void assertAnnouncedToThread( MockThreadListener *thread, Event &event )
    {
        TS_ASSERT( thread->announceWasCalled );
        TS_ASSERT_EQUALS( thread->lastEventType, event.type() );

        thread->announceWasCalled = false;
    }

    void assertCoreMessageSent( unsigned long long index, const String &label )
    {
        TS_ASSERT( engine->coreMessageSent );
        TS_ASSERT_EQUALS( engine->lastSentMessageIndex, index );
        TS_ASSERT_EQUALS( engine->lastSentLabel, label );

        engine->coreMessageSent = false;
    }

    void receiveCoreMessage( Weaver &weaver, unsigned long long index, const String &label )
    {
        CoreMessage coreMessage( index, "bla", label );

        TS_ASSERT_THROWS_NOTHING( weaver.incomingCoreMessage( coreMessage ) );
    }

    void assertNoCoreMessageSent()
    {
        TS_ASSERT( !engine->coreMessageSent );
    }

    Vector<Event> dontCare;

    void xtest_wait_for_non_determinism()
    {
        Weaver weaver( *selector, *engine );

        setThreadName( socket1, "Jefferson" );

        Event event1( 1 );
        Event event2( 2 );

        ThreadStateGraph threadStateGraphOne;
        prepareStateGraphOne( threadStateGraphOne );
        TS_ASSERT_THROWS_NOTHING( weaver.addLocalStateGraph( "Jefferson", threadStateGraphOne ) );

        ThreadStateGraph threadStateGraphTwo;
        prepareStateGraphOne( threadStateGraphTwo );
        TS_ASSERT_THROWS_NOTHING( weaver.addForeignStateGraph( "Darcy", threadStateGraphTwo ) );

        TS_ASSERT_THROWS_NOTHING( weaver.handleAcceptedSocket( *socket1 ) );

        assertAnnouncedToThread( threadListener1, event1 );
        assertNoCoreMessageSent();

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( threadListener1, dontCare, dontCare, dontCare, "b", 12 ) );

        assertCoreMessageSent( 1, "b" );
        receiveCoreMessage( weaver, 1, "d" );

        assertAnnouncedToThread( threadListener1, event1 );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( threadListener1, dontCare, dontCare, dontCare, "a", 13 ) );

        assertAnnouncedToThread( threadListener1, event1 );
        assertCoreMessageSent( 3, "a" );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
